#! /usr/bin/python3 -O
#
# jBr - A set of tools to manage your photo library.
# Copyright (C) 2023-2024 MiKael NAVARRO
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""jBr CLI - A set of tools to manage your photo library.
Copyright (C) 2023-2024 MiKael NAVARRO

A collection of command line tools to manage your photo albums in the spirit of jBrout (https://jbrout.manatlan.com).

"""

# Specific variables for pydoc
__author__ = "MiKael Navarro <klnavarro@mailo.com>"
__date__ = "Wed 3 Jan 2024"
__version__ = "0.1.1"
__credits__ = """Thanks to Marc Lentz (aka manatlan) for his photo management software jBrout (https://jbrout.manatlan.com) which is my source of inspiration."""


# Include directives
import sys
import os
import shutil
import pathlib
import glob
from pprint import pprint as pp, pformat
from datetime import datetime
from PIL import Image, ImageTk
from PIL.ExifTags import TAGS
from PIL.ImageOps import exif_transpose
from pillow_heif import register_heif_opener
import exifread
from iptcinfo3 import IPTCInfo
import pyexiv2  # need boost.python3
import tkinter as tk
from tkinter import ttk
from tkinter.simpledialog import askstring
from tkinter.filedialog import askopenfilenames
from tkinter.messagebox import showinfo
from tkinter.scrolledtext import ScrolledText


# Global variables
TIMESTAMP_FORMAT = "p%Y%m%d_%H%M%S"
TIMESTAMP_ISO8601 = "%Y%m%dT%H%M%S.%f"
EXIF_WIDTH = 256
EXIF_HEIGHT = 257
EXIF_BITS_PER_SAMPLE = 258
EXIF_DESCRIPTION = 270
EXIF_ORIENTATION = 274
EXIF_X_RESOLUTION = 282
EXIF_Y_RESOLUTION = 283
EXIF_RESOLUTION_UNIT = 296
EXIF_SOFTWARE = 305
EXIF_DATETIME = 306
IPTC_KEYWORDS = 25
IPTC_DATE_CREATED = 55
IPTC_TIME_CREATED = 60


#
# Print to STDERR.
#
def eprint(*args, **kwargs):
    print("%s: error:" % os.path.basename(__file__), file=sys.stderr, end=" ")
    print(*args, file=sys.stderr, **kwargs)


#
# Metadata.
#
def metadata(files, full=False):
    """Return Exif/IPTC info from given tags."""

    ret = {}
    print("Analyzing '%s' files..." % files)

    # Loop on all files:
    for pattern in files:
        for infile in glob.glob(pattern):
            filename, ext = os.path.splitext(infile)
            dirname = os.path.dirname(filename)

            #print("%s/%s%s:" % (dirname, os.path.basename(filename), ext))
            ret[infile] = {}
            info = ret[infile]

            # PIL EXIF
            with Image.open(infile) as img:
                exif = img.getexif()
                info['PIL EXIF'] = {TAGS.get(t, t): v for t, v in exif.items()}

            # ExifRead
            if full:
                with open(infile, 'rb') as fd:
                    info['ExifRead'] = exifread.process_file(fd)

            # IPTCInfo3
            iptcinfo = IPTCInfo(infile, force=True, inp_charset='utf-8')
            info['IPTCInfo3'] = {iptcinfo._data._key_as_str(k): v for k, v in iptcinfo._data.items()}

            # pyExiv2
            if full:
                metadata = pyexiv2.ImageMetadata(infile)
                metadata.read()
                info['pyExiv2'] = {}
                info['pyExiv2'] = info['pyExiv2'] | {k: metadata[k].raw_value for k in metadata.exif_keys}
                info['pyExiv2'] = info['pyExiv2'] | {k: metadata[k].raw_value for k in metadata.iptc_keys}
                info['pyExiv2'] = info['pyExiv2'] | {k: metadata[k].raw_value for k in metadata.xmp_keys}
                info['pyExiv2'] = info['pyExiv2'] | {'previews': [p.dimensions, p.mime_type, p.size] for p in metadata.previews}

    return ret


#
# Add/remove keywords.
#
def keywords(files, keywords_add=None, keywords_del=None):
    """Add/remove kewwords to given files."""

    keywords_add = keywords_add and set(keywords_add) or set()
    keywords_del = keywords_del and set(keywords_del) or set()

    ret = {}

    # Loop on all files:
    for pattern in files:
        for infile in glob.glob(pattern):
            filename, ext = os.path.splitext(infile)
            dirname = os.path.dirname(filename)

            ret[infile] = {}
            info = ret[infile]

            # IPTCInfo3
            iptcinfo = IPTCInfo(infile, force=True, inp_charset='utf-8')

            #if not iptcinfo:
            #    eprint("no IPTC info, please check")
            #    continue

            if keywords_add:
                print("Add %s keywords to '%s'..." % (keywords_add, infile))
                iptcinfo['keywords'] = [k for k in set(iptcinfo['keywords']) | keywords_add]

            if keywords_del:
                print("Remove %s keywords from '%s'..." % (keywords_del, infile))
                iptcinfo['keywords'] = [k for k in iptcinfo['keywords'] if k not in keywords_del]

            #print("%s/%s%s:" % (dirname, os.path.basename(filename), ext))

            info['keywords'] = iptcinfo['keywords']

            if keywords_add or keywords_del:
                iptcinfo.save()
                iptcinfo.save_as(infile, options=['overwrite'])

    return ret


#
# Filter by tags.
#
def filter(files, tags=None):
    """Filter fites accordingly to given tags."""

    tags = tags and set(tags) or set()

    print("Filtering '%s' files by %s..." % (files, tags))
    res = []

    # Loop on all files:
    for pattern in files:
        for infile in glob.glob(pattern):
            filename, ext = os.path.splitext(infile)

            iptcinfo = IPTCInfo(infile, force=True)
            keys = {k.decode(encoding='ascii') for k in iptcinfo['keywords']}

            # All tags must be in keys
            if (keys & tags) == tags:
                res.append(infile)

    return res


#
# Horodatage.
#
def timestamp(files, override=True):
    """Rename a photo based on when they were taken."""

    nb = 0

    # Loop on all files:
    for pattern in files:
        for infile in glob.glob(pattern):
            filename, ext = os.path.splitext(infile)
            dirname = os.path.dirname(filename)

            # Open file (with PIL)
            with Image.open(infile) as img:
                if img.format not in ["JPEG"]:
                    eprint("bad image format '%s', must be 'JPEG'" % img.format)
                    continue

                exif = img.getexif()  # to get timestamp

                dt = datetime.now()
                try:
                    dt = datetime.strptime(exif[EXIF_DATETIME], "%Y:%m:%d %H:%M:%S")
                except KeyError:  # missing 'DateTime' key, defaulting to file timestamp
                    dt = datetime.fromtimestamp(os.path.getmtime(infile))

                fn = "%s/%s.%s" % (dirname, dt.strftime(TIMESTAMP_FORMAT), "jpg")
                print("Renaming '%s' to '%s'..." % (img.filename, fn))

                if override:
                    os.replace(infile, fn)
                else:
                    try:
                        shutil.copy2(infile, fn)
                    except shutil.SameFileError as e:
                        eprint(e)

            nb += 1

    return nb  # number of files processed


#
# Strip metadata.
#
def strip(files):
    """Remove all EXIF/IPTC/XMP metadata from given files."""

    nb = 0

    # Loop on all files:
    for pattern in files:
        for infile in glob.glob(pattern):
            filename, ext = os.path.splitext(infile)
            dirname = os.path.dirname(filename)

            print("Stripping Exif metadata from '%s/%s%s'..." % (dirname, os.path.basename(filename), ext))

            # Open file (with PIL)
            with Image.open(infile) as img:
                if img.format not in ["JPEG"]:
                    eprint("bad image format '%s', must be 'JPEG'" % img.format)
                    continue

                img.getexif().clear()

                # Overwrite!!!
                img.save(infile, format=img.format.lower)

            nb += 1

    return nb  # number of files processed


#
# Convert file format.
#
def convert(files, archive=False):
    """Convert files to jpeg format."""

    nb = 0

    # Loop on all files:
    for pattern in files:
        for infile in glob.glob(pattern):
            filename, ext = os.path.splitext(infile)
            dirname = os.path.dirname(filename)

            fn = "%s/%s.%s" % (dirname, os.path.basename(filename), "jpg")
            print("Converting '%s' to '%s'..." % (infile, fn))

            # Open file (with PIL)
            with Image.open(infile) as img:
                if img.format not in ["HEIC", "HEIF"]:
                    eprint("bad image format '%s', must be 'HEIC' or 'HEIF'" % img.format)
                    continue

                img.save(fn, format="JPEG", quality=90, exif=img.getexif())

                if archive:
                    dn = "%s/heic/" % dirname
                    os.makedirs(dn, mode=0o755, exist_ok=True)
                    shutil.move(infile, dn)
                else:  # remove
                    fn = pathlib.Path(infile)
                    fn.unlink(missing_ok=True)

            nb += 1

    return nb  # number of files processed


#
# Recalc. thumbnail.
#
def thumbnail(files):
    """Re-create thumbnail of given files based on EXIF metadata."""

    nb = 0

    # Loop on all files:
    for pattern in files:
        for infile in glob.glob(pattern):
            filename, ext = os.path.splitext(infile)
            dirname = os.path.dirname(filename)

            print("Re-create thumbnail of '%s/%s%s'..." % (dirname, os.path.basename(filename), ext))

            # Open file (with PIL)
            with Image.open(infile, mode='r') as img:
                if img.format not in ["JPEG"]:
                    eprint("bad image format '%s', must be 'JPEG'" % img.format)
                    continue

                dn = "%s/thumb/" % dirname
                os.makedirs(dn, mode=0o755, exist_ok=True)
                tn = "%s/%s%s" % (dn, os.path.basename(filename), ext)

                # Create thumbnail
                img.thumbnail((128, 128))
                img.save(tn, "JPEG")

                # Replace thumbnail EXIF metadata (JPEGInterchangeFormat)
                mt = pyexiv2.ImageMetadata(infile)
                mt.read()
                th = mt.exif_thumbnail
                print([[p.dimensions, p.mime_type, p.size] for p in mt.previews])
                th.set_from_file(tn)
                print([[p.dimensions, p.mime_type, p.size] for p in mt.previews])
                mt.write()

            nb += 1

    return nb  # number of files processed


#
# Correct orientation.
#
def transpose(files):
    """Correct orientation of given files accordingly to EXIF metadata."""

    nb = 0

    # Loop on all files:
    for pattern in files:
        for infile in glob.glob(pattern):
            filename, ext = os.path.splitext(infile)
            dirname = os.path.dirname(filename)

            print("Correct orientation of '%s/%s%s'..." % (dirname, os.path.basename(filename), ext))

            # Open file (with PIL)
            with Image.open(infile) as img:
                if img.format not in ["JPEG"]:
                    eprint("bad image format '%s', must be 'JPEG'" % img.format)
                    continue

                exif_transpose(img, in_place=True)

            nb += 1

    return nb  # number of files processed


#
# Simple message dialog
#
def showmsg(title="Information...", message=None):
    """Display message window."""

    root = tk.Tk()
    root.title(title)
    root.geometry("%dx%s" % (130 * 4 + 20, 300))  # 4 columns
    root.resizable(True, True)

    text = ScrolledText(root, wrap=tk.WORD)
    text.pack(fill=tk.BOTH, expand=True)

    text.insert("1.0", message)
    text.focus()

    root.mainloop()


#
# Simple photo previews
#
def showimg(title="Preview photos...", files=[]):
    """Display thumbnails in scrollable frame + show on click."""

    nb_cols = 5
    th_size = (128, 128)

    root = tk.Tk()
    root.title(title)
    root.geometry("%dx%s" % ((th_size[0] + 2) * nb_cols + 20, 300))  # nb columns
    root.resizable(False, True)

    frame = ttk.Frame(root)  # main frame
    frame.pack(fill=tk.BOTH, expand=True)

    canvas = tk.Canvas(frame)

    scrollbar = ttk.Scrollbar(frame, orient=tk.VERTICAL, command=canvas.yview)
    scrollable_frame = ttk.Frame(canvas)  # scrollable frame that contains labels

    scrollable_frame.bind("<Configure>", lambda e: canvas.configure(scrollregion=canvas.bbox("all")))

    canvas.create_window((0, 0), window=scrollable_frame, anchor=tk.NW)
    canvas.configure(yscrollcommand=scrollbar.set)

    row = col = 0
    for infile in files:
        with Image.open(infile) as img:
            img.thumbnail(th_size)
            img = ImageTk.PhotoImage(img)

            label = ttk.Label(scrollable_frame, image=img, compound=tk.TOP,
                              text=os.path.basename(infile))
            label.photo = img  # assign to class variable to resolve problem with bug in `PhotoImage`
            label.bind("<Button-1>", lambda e, f=infile: Image.open(f).show())
            label.grid(row=row, column=col)

            col = (col + 1) % nb_cols  # next row/col
            row += (col == 0) and 1 or 0

    canvas.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
    scrollbar.pack(side=tk.RIGHT, fill=tk.Y)

    root.mainloop()


#
# Main entry point.
#
def main(args):
    """Perform commands..."""

    #print(args)

    register_heif_opener()

    if hasattr(args, "metadata_files"):  # display EXIF
        metadata_files = args.metadata_files
        full = args.full

        if args.tkinter:
            if not metadata_files:
                metadata_files = askopenfilenames(title="Select files to examine...",
                                                  filetypes=(("JPEG", ("*.jpg", "*.jpeg")),
                                                             ("All", "*")))
            full = False

        if metadata_files:
            ret = metadata(list(metadata_files), full)
            pp(ret)

            if args.tkinter:
                showmsg(title="Metadata (EXIF/IPTC)...", message=pformat(ret))

    elif hasattr(args, "filter_files"):  # filter by tags
        filter_files = args.filter_files
        filter_tags = args.filter_tags and args.filter_tags or []

        if args.tkinter:
            if not filter_files:
                filter_files = askopenfilenames(title="Select files to filter by tags...",
                                                filetypes=(("JPEG", ("*.jpg", "*.jpeg")),
                                                           ("All", "*")))

            filter_tags = askstring(title="Filter tags",
                                    prompt="Tags (comma separated)",
                                    initialvalue=",".join(filter_tags))
            if filter_tags:
                filter_tags = [t.strip() for t in filter_tags.split(',')]

        if filter_files:
            res = filter(filter_files, filter_tags)
            pp(res)

            if args.tkinter:
                showimg(title="Filtered photos...", files=res)

    elif hasattr(args, "keywords_files"):  # add/del keywords
        keywords_files = args.keywords_files
        keywords_add = args.keywords_add and args.keywords_add or []
        keywords_del = args.keywords_del and args.keywords_del or []

        if args.tkinter:
            if not keywords_files:
                keywords_files = askopenfilenames(title="Select files to add/del keywords...",
                                                  filetypes=(("JPEG", ("*.jpg", "*.jpeg")),
                                                             ("All", "*")))

            keywords_add = askstring(title="Add keywords",
                                     prompt="Keywords (comma separated)",
                                     initialvalue=",".join(keywords_add))
            if keywords_add:
                keywords_add = [k.strip() for k in keywords_add.split(',')]

            keywords_del = askstring(title="Del keywords",
                                     prompt="Keywords (comma separated)",
                                     initialvalue=",".join(keywords_del))
            if keywords_del:
                keywords_del = [k.strip() for k in keywords_del.split(',')]

        if keywords_files:
            ret = keywords(keywords_files, keywords_add, keywords_del)
            pp(ret)

            if args.tkinter:
                showmsg(title="Keywords(IPTC)...",
                        message=pformat({f: v['keywords'] for f, v in ret.items()}))

    elif hasattr(args, "rename_files"):  # timestamp
        rename_files = args.rename_files
        override = (not args.keep)

        if args.tkinter:
            if not rename_files:
                rename_files = askopenfilenames(title="Select files to rename (timestamp)...",
                                                filetypes=(("JPEG", ("*.jpg", "*.jpeg")),
                                                           ("All", "*")))
            override = False

        if rename_files:
            nb = timestamp(rename_files, override)

            if args.tkinter:
                showinfo(title="Timestamp photos...", message="%d photos processed." % nb)

    elif hasattr(args, "strip_files"):  # strip EXIF
        strip_files = args.strip_files

        if args.tkinter:
            if not strip_files:
                strip_files = askopenfilenames(title="Select files to strip metadata (EXIF)...",
                                               filetypes=(("JPEG", ("*.jpg", "*.jpeg")),
                                                          ("All", "*")))

        if strip_files:
            nb = strip(strip_files)

            if args.tkinter:
                showinfo(title="Stripping EXIF...", message="%d photos processed." % nb)

    elif hasattr(args, "convert_files"):  # HEIC -> JPEG
        convert_files = args.convert_files
        archive = args.archive

        if args.tkinter:
            if not convert_files:
                convert_files = askopenfilenames(title="Select files to convert (HEIC -> JPEG)...",
                                                 filetypes=(("HEIC", "*.heic"),
                                                            ("All", "*")))
            archive = True

        if convert_files:
            nb = convert(convert_files, archive=archive)

            if args.tkinter:
                showinfo(title="Converting photos...", message="%d photos processed." % nb)

    elif hasattr(args, "thumbnail_files"):  # re-create thumbnail
        thumbnail_files = args.thumbnail_files

        if args.tkinter:
            if not thumbnail_files:
                thumbnail_files = askopenfilenames(title="Select files to re-create thumbnail...",
                                                   filetypes=(("JPEG", ("*.jpg", "*.jpeg")),
                                                              ("All", "*")))

        if thumbnail_files:
            nb = thumbnail(thumbnail_files)

            if args.tkinter:
                showinfo(title="Creating thumbnails...", message="%d photos processed." % nb)

    elif hasattr(args, "transpose_files"):  # correct orientation
        transpose_files = args.transpose_files

        if args.tkinter:
            if not transpose_files:
                transpose_files = askopenfilenames(title="Select files to correct orientation...",
                                                   filetypes=(("JPEG", ("*.jpg", "*.jpeg")),
                                                              ("All", "*")))

        if transpose_files:
            nb = transpose(transpose_files)

            if args.tkinter:
                showinfo(title="Correcting orientation...", message="%d photos processed." % nb)

    else:
        eprint("command missing (see --help)")


#
# External entry point.
#
if __name__ == "__main__":
    # Parsing command line (http://www.doughellmann.com/PyMOTW/argparse/)
    import argparse

    parser = argparse.ArgumentParser(description="A set of tools to manage your photo library.",
                                     prefix_chars='-+')
    parser.add_argument('-V', '--version',
                        action='version', version=__version__)

    subparsers = parser.add_subparsers(help='commands')

    # Metadata command
    metadata_parser = subparsers.add_parser('metadata',
                                            help="display metadata (EXIF/IPTC) of given files")
    metadata_parser.add_argument('metadata_files', nargs='*',
                                 help="unix style filename pattern (ex. \"*.jpeg\" *.jpg)")
    metadata_parser.add_argument('-f', '--full', action='store_true',
                                 default=False,
                                 help="display all metadata (EXIF/IPTC/XMP)")

    # Filter command
    filter_parser = subparsers.add_parser('filter', prefix_chars='-+',
                                          help="filter files accordingly to their tags")
    filter_parser.add_argument('filter_files', nargs='*',
                               help="unix style filename pattern (ex. \"*.jpeg\" *.jpg)")
    filter_parser.add_argument('+f', '--tags', action='append',
                               dest='filter_tags', default=None,
                               help="tags to filter (ex. +k avion +k \"salon de provence\")")

    # Keywords command
    keywords_parser = subparsers.add_parser('keywords', prefix_chars='-+',
                                            help="add/remove keywords to/from given files")
    keywords_parser.add_argument('keywords_files', nargs='*',
                                 help="unix style filename pattern (ex. \"*.jpeg\" *.jpg)")
    keywords_parser.add_argument('+k', '--add', action='append',
                                 dest='keywords_add',
                                 help="keywords to add (ex. +k meeting +k \"salon de provence\")")
    keywords_parser.add_argument('-k', '--delete', action='append',
                                 dest='keywords_del',
                                 help="keywords to remove (ex. -k aix -k salon)")

    # Rename command
    rename_parser = subparsers.add_parser('timestamp',
                                          help="rename files accordingly to their timestamp")
    rename_parser.add_argument('rename_files', nargs='*',
                               help="unix style filename pattern (ex. \"*.jpeg\" *.jpg)")
    rename_parser.add_argument('-k', '--keep', action='store_true',
                               default=False,
                               help="keep original files (do not remove)")

    # Strip command
    strip_parser = subparsers.add_parser('strip',
                                         help="strip metadata (EXIF) of given files")
    strip_parser.add_argument('strip_files', nargs='*',
                              help="unix style filename pattern (ex. \"*.jpeg\" *.jpg)")

    # Convert command
    convert_parser = subparsers.add_parser('convert',
                                           help="convert given files from HEIC to JPEG")
    convert_parser.add_argument('convert_files', nargs='*',
                                help="unix style filename pattern (ex. \"*/*.heic\" *.heic)")
    convert_parser.add_argument('-a', '--archive', action='store_true',
                                default=False,
                                help="archive original files (into 'heic/' subdirectory)")

    # Thumbnail command
    thumbnail_parser = subparsers.add_parser('thumbnail',
                                             help="re-create thumbnail (EXIF) of given files")
    thumbnail_parser.add_argument('thumbnail_files', nargs='*',
                                  help="unix style filename pattern (ex. \"*.jpeg\" *.jpg)")

    # Transpose command
    transpose_parser = subparsers.add_parser('transpose',
                                             help="correct orientation accordingly to EXIF metadata")
    transpose_parser.add_argument('transpose_files', nargs='*',
                                  help="unix style filename pattern (ex. \"*.jpeg\" *.jpg)")

    # Tkinter
    parser.add_argument('--tkinter', action='store_true',
                        default=False,
                        help="ask Tkinter message dialogs")

    # Parse!
    args = parser.parse_args()

    # Process start here
    main(args)
